/*--------------------------------------------------------
*      AVR-Basic-Interpreter mit SD-Karte
*    =====================================
*     Uwe Berger (bergeruw@gmx.net); 2010
* 
* Basic-Programme liegen auf einer SD-Karte. Fuer die SD-
* Kartenzugriffe wurde folgende Lib verwendet:
*
* http://www.mikrocontroller.net/articles/AVR_FAT32
*
*
* Die "Shell-Befehle" wurden moeglichst einfach gehalten:
*
* m             : Speicherinformationen (SRAM)
* d             : Dateien auf SD-Karte auflisten
* l <Dateiname> : Programmlisting
* r <Dateiname> : Programm starten
* t             : Runtime ausgeben
*
* Dateinamen sind im Format 8.3; sind aber ohne Punkt und vollen 8
* stellen anzugeben; wie bei Kommando d angezeigt. Bsp.:
*
*               12345678901
*               vvvvvvvvvvv
* test1.bas --> test1   bas
*
* Gross-/Kleinschreibung der Kommandos/Dateinamen ist egal.
*
* Die Dateien muessen im Root-Verzeichnis der SD-Karte liegen.
* 
* Ein-/Ausgaben erfolgen ueber die serielle Schnittstelle (57600, 8, n, 1)
*
*
* Hinweis: es laufen nicht mehr alle Basic-Beispiele, wenn man einen
* ATmega168 mit 1kByte SRAM verwendet, da man langsam der dynamische
* Speicher ausgeht. Abhilfe z.B. einen ATmega328 mit 2 kByte SRAM ver-
* wenden....
*
*
* ---------
* Have fun!
*
----------------------------------------------------------*/

#include <stdlib.h>
#include <avr/io.h>

#include "sd_card/config.h"
#include "sd_card/file.h"
#include "sd_card/fat.h"
#include "sd_card/mmc.h"
#include "mem_check/mem_check.h"

#include "uart/usart.h"
#include "avr_basic/tokenizer_access.h"
#include "avr_basic/ubasic.h"

//Taktfrequenz
#ifndef F_CPU
#define F_CPU 16000000UL
#endif

//Baudrate der seriellen Schnittstelle
#define BAUDRATE 57600

#define PRELOAD_T0 6

volatile uint16_t time_row;
uint16_t time_all;

//************************************************************************
ISR(TIMER0_OVF_vect)
{
	TCNT0 = PRELOAD_T0;						// Preloadwert setzen
	time_row++;
}


//*****************************************************************************************************************
int main(void){
	char buffer[15];


    usart_init(BAUDRATE);

    // Timer 0 initialisieren (runtime-Messung)
    TCNT0  = PRELOAD_T0;							// Preloadwert Timer0 setzen
	TCCR0B |= (1 << CS01)|(1 << CS00);				// Prescaler Timer0 auf 64 stellen
	TIMSK0 |= (1 << TOIE0);
	sei();
	

	usart_write("AVR-Basic-Computer; Uwe Berger, 2010\r\n");
	usart_write("Compiliert am "__DATE__" um "__TIME__"\r\n");

	
	// MMC-INIT...	
	while (FALSE==mmc_init()){
		nop();
	}
	usart_write("SD-Card ok...\r\n");


	// FAT-INIT
	while (FALSE == fat_loadFatData()){
		nop();
	}
	usart_write("FAT ok...\n\r");

	usart_write("\r\nunused mem: %i\r\n", get_mem_unused());
	usart_write("free mem..: %i\r\n", get_mem_free());
    usart_write("\r\n>");


	while(1){
		char* command = buffer;
        if(usart_read_line(command, sizeof(buffer)) < 1) continue;
        for (uint8_t i=0; i<strlen(buffer); i++) buffer[i] = toupper(buffer[i]);
        // Speicherinformationen (SRAM)
        if (strcmp_P(command, PSTR("M")) == 0) {
			usart_write("\r\nunused mem: %i\r\n", get_mem_unused());
			usart_write("free mem..: %i\r\n", get_mem_free());
		} else
		// Root-Verzeichnisinhalt auflisten
		if (strcmp_P(command, PSTR("D")) == 0) {
			ffls((fptr)usart_write_str);
		} else
        // Basic-Programm ausfuehren
        if (strncmp_P(command, PSTR("R "), 2) == 0) {
			usart_write("\r\n");
			command += 2;
			buffer[14]=0;
			if (ffileExsists((unsigned char*)command)) {
				ffopen((unsigned char*)command);
				time_all=0;
				ubasic_init(0);
				do {
					time_row=0;
					ubasic_run();
					time_all += time_row;
					if (usart_is_receive() && (usart_receive_char() == 0x1b)) {
						usart_write("\r\nProgramm unterbrochen...!\r\n");
						break;
					}
				} while(!ubasic_finished());
				ffclose();
			} else {
				usart_write("Fehler beim Oeffnen der Datei: %s\n\r", command);
			}
		} else
		// Runtime ausgeben
        if (strcmp_P(command, PSTR("T")) == 0) {
			usart_write("\r\nruntime: %ims\r\n", time_all);
        } else
        // Basic-Programm auflisten
        if (strncmp_P(command, PSTR("L "), 2) == 0) {
			usart_write("\r\n");
			command += 2;
			buffer[14]=0;
			if (ffileExsists((unsigned char*)command)) {
				ffopen((unsigned char*)command);
				// Dateilaenge setzen
				unsigned long int seek=file.length;
				char c;
				do {
					c=ffread();
					usart_write_char(c);
					if (c=='\n') usart_write_char('\r');
				} while(--seek);
				ffclose();
			} else {
				usart_write("Fehler beim Oeffnen der Datei: %s\n\r", command);
			}
        } else 	usart_write("\r\nKommando %s unbekannt!\r\n", command);
        // Prompt
        usart_write("\r\n>");
    }
	return 0;
}
